/*
** Module   : ProcessDaemon.java
** Abstract : process launching server
**
** Copyright (c) 2006-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description---------------------
** 001 GES 20060212   @24563 First version supporting 2 forms: a launch
**                           that has its STDIO connected to the terminal
**                           and a launch which has STDIO redirected to
**                           and/or from streams that are accessible from
**                           the caller.
** 002 GES 20060213   @24587 Changes to use custom process launching
**                           path (bypassed the J2SE launching) when
**                           in non-streams mode. This is required since
**                           the J2SE Runtime.exec() does not provide a
**                           terminal to the child process.  This makes
**                           it impossible to handle interactive child
**                           processes.
** 003 GES 20060518   @26343 Changes to use custom process launching
**                           path (bypassed the J2SE launching) when
**                           in streams mode since Progress provides a
**                           terminal to the child process.
** 004 GES 20060530   @26678 Rethrow ConditionException instances during
**                           launch (can occur due to end user input in
**                           pause processing).
** 005 GES 20061020   @30578 Return lastkey from interactive forms of
**                           launch since any pause that occurs during
**                           that time will update lastkey.
** 006 NVS 20070308   @32294 Process launching has been reworked and is no
**                           longer based on providing a terminal device
**                           to the child process through pseudo-terminal.
**                           Instead, the parent process' TTY device is
**                           directly used by any stdio handle which is
**                           not redirected to a pipe.
** 007 NVS 20070309   @32344 Some minor cleanup.
** 008 GES 20080123   @36899 Removed dead code.
** 009 NVS 20080212   @37035 Minor changes required. Process launching and
**                           waiting for the child now take a flag telling
**                           the process is interactive.
** 010 SIY 20090622   @42774 Detect special "silent" mode, when invalid 
**                           external application is just ignored and pipe
**                           assumed live despite the fact that external 
**                           application even was not started.
** 011 GES 20111006          Directly use launch mgr for pseudoTerminal methods
**                           instead of using thin client.
** 012 GES 20111025          Modified API export code to capture and use the
**                           authentication token so that multiple registrations
**                           can be done per JVM session. This occurs when the
**                           client code is running in the server process and
**                           there is a CTRL-C that causes the client to
**                           restart.
** 013 GES 20111111          Modified interface export to allow registration
**                           to be exclusively local or remote as needed.
** 014 CS  20121213          Added support for calling the Quoter utility.
*/


package com.goldencode.p2j.util;

import java.io.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.chui.*;

/**
 * Provides 2 forms of process launching: a launch that has its STDIO 
 * connected to the terminal and a launch which has STDIO redirected to
 * and/or from streams that are accessible from the caller.
 * <p>
 * This is designed to launch processes on a remote network node. 
 */
public class ProcessDaemon
implements Launcher
{   
   /** Token used to authenticate with the dispatcher when registering APIs. */
   private static Object modToken = null;

   /** Represents the terminal. */
   private ThinClient terminal = null;
      
   /** Stream daemon (provides access to streams by id). */
   private StreamDaemon sd = null;
   
   /**
    * Initializes the process server.
    *
    * @param    terminal
    *           Provides access to terminal features and helpers.
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    * @param    sd
    *           Provides translation from id to stream instance.
    */
   public ProcessDaemon(ThinClient terminal, boolean single, StreamDaemon sd)
   {
      this.terminal = terminal;
      this.sd       = sd;
      
      synchronized (ProcessDaemon.class)
      {
         modToken = RemoteObject.registerServer(Launcher.class,
                                                this,
                                                modToken,
                                                single);
      }
   }
   
   /**
    * Launches the external executable specified by the <code>cmdlist</code>
    * taking into account the flags which specify synchronous or asynchronous
    * operation and whether a prompt is provided after the child process
    * terminates. The command is always executed as a child process of the 
    * shell which is a child process of the JVM process on the node in which
    * this is served. 
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> of the
    * child process refer to the terminal this JVM is running on.
    * <p>
    * Since the <code>Runtime.exec()</code> process launching method of J2SE
    * does not provide a terminal/console to the launched child process,
    * such a method cannot be used with child processes that must integrate
    * with the terminal/console of the current process.  In particular, it
    * is impossible to use interactive child processes with the J2SE method
    * of process launching.  For this reason, an alternate process launching
    * infrastructure is used in <code>ThinClient.pseudoTerminalLaunch</code>.
    * This method is used to start a process such that it does have a
    * terminal.  When used with the current process' standard I/O streams,
    * this is an effective method of launching interactive child processes.
    * All proper wait and pause semantics are honored and the current process'
    * terminal processing is properly suspended and resumed as needed.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    silent
    *           If <code>true</code> the default prompt is suppressed.
    *           Otherwise, {@link ThinClient#pause} is called when the
    *           child process exits.
    * @param    wait
    *           If <code>true</code>, synchronous operation is requested
    *           which waits for the termination of the child process on the
    *           current thread.  If <code>false</code>, a thread will be
    *           started to wait for the child process termination and to
    *           subsequently handle cleanup (asynchronous mode).
    *
    * @return   The key code for any key pressed during any pause or -1 if
    *           there was no pause.
    */
   public int launch(String[] cmdlist, boolean silent, boolean wait)
   {
      int lastkey = -1;
      
      try
      {         
         // clear the thin client's frame output and suspend normal UI mode          
         terminal.suspend();
         
         // preprocess the parms
         String[] list = prepareCommandLine(cmdlist);
         
         // launch the process
         int pid = LaunchManager.pseudoTerminalLaunch(list, null, null, null, 
                                                      silent);
         
         if (pid < 0)
         {
            // re-initialize the terminal
            terminal.resume();
            
            ErrorManager.recordOrThrowError(pid, "Process launch failed.");
            return lastkey;
         }
         
         // synch or asynch?
         if (wait)
         {
            // block until the child process exits
            LaunchManager.pseudoTerminalWait(pid, silent);
         }
         else
         {
            // create a thread to handle cleanup asynchronously
            Thread t1 = new Thread(new Cleaner(pid));
            t1.start();
         }
         
         // if not silent we display a default message and pause indefinitely
         if (!silent)
         {
            lastkey = terminal.suspendedPause();
         }         
      }
      
      catch (ConditionException ce)
      {
         // rethrow any condition exceptions (could occur during pause 
         // processing on END-ERROR... for example)
         throw ce;
      }
      
      catch (Exception exc)
      {
         // silent failure like Progress...
      }      
      
      finally
      {
         // resume normal operations and clear any child process output
         terminal.resume();         
      }
      
      return lastkey;
   }
   
   /**
    * Launches the external executable specified by the <code>cmdlist</code>
    * on an asynchronous basis. The command is always executed as a child
    * process of the shell which is a child process of the the JVM process on
    * the node in which this is served. This corresponds to the standard
    * concept of "co-processing" where the parent process and child process
    * are both running at the same time and they interact in one direction
    * or both directions by reading and/or writing from connected pipes.
    * <p>
    * <code>STDIN</code>, <code>STDOUT</code> and <code>STDERR</code> are
    * all linked from the child process to the passed parameters if
    * those parameters are not <code>null</code>. Any <code>null</code>
    * pipes are closed immediately and those pipes connected with the
    * passed parameters must be closed by the caller.  Please see the
    * {@link ProcessStream} for more details.
    *
    * @param    cmdlist
    *           The command line to be executed, with the first element
    *           used as the program to run and all other array elements as
    *           the arguments.
    * @param    sout
    *           The id of a {@link ProcessStream} instance to be connected to  
    *           the child process' <code>STDOUT and STDERR</code> or
    *           -1 for no connection.  Used for reading.
    * @param    sin
    *           The id of a <code>ProcessStream</code> instance to be
    *           connected to the child process' <code>STDIN</code> or
    *           -1 for no connection.  Used for writing.
    */
   public void launch(String[] cmdlist, int sout, int sin)
   {
      try
      {     
         boolean broken = false;
         
         // create file descriptors
         FileDescriptor fdIn  = sin  != -1 ? new FileDescriptor() : null;
         FileDescriptor fdOut = sout != -1 ? new FileDescriptor() : null;
         FileDescriptor fdErr = sout != -1 ? new FileDescriptor() : null;
         
         if (cmdlist != null && 
             cmdlist.length == 1 && 
             cmdlist[0].equals("?") && 
             fdIn != null)
         {
            // special case of silent error mode
            broken = true; 
         }
         
         // preprocess the parms
         String[] list = prepareCommandLine(cmdlist);
         
         // launch the process
         int pid = LaunchManager.pseudoTerminalLaunch(list, fdIn, fdOut, fdErr, 
                                                      true);
         
         if (pid < 0)
         {
            ErrorManager.recordOrThrowError(pid, "Process launch failed.");
            return;
         }
         
         // create the stdio of the child process
         OutputStream cin = fdIn  != null ? new FileOutputStream(fdIn) : null;
         InputStream cout = fdOut != null ? new FileInputStream(fdOut) : null;
         InputStream cerr = fdErr != null ? new FileInputStream(fdErr) : null;
         
         // if the corresponding stream argument is not null, create pipes to
         // copy the stdout/stderr output from the child process to our
         // stream arguments, otherwise just close the child process' pipe
         if (sout != -1)
         {
            ProcessStream pso = (ProcessStream) sd.streamFromId(sout);
            pso.connectOut(pid, cout, cerr);
         }         
         
         if (sin != -1)
         {
            ProcessStream psi = (ProcessStream) sd.streamFromId(sin);
            psi.connectIn(pid, cin);
            
            if (broken)
               psi.setBroken();
         }
      }
      
      catch (Exception exc)
      {
         // silent failure like Progress...
      }            
   }
   
   /**
    * Converts the given command list into a form that is suitable for
    * process launching.  In particular, it creates an array with the
    * first element set to the system shell and subsequent elements are
    * shell options and the reconstituted command line from the original
    * caller (this is built up from the passed command list, back into a
    * single string).
    *
    * @param    cmdlist
    *           The command to execute (split into whitespace delimited
    *           pieces.
    *
    * @return   An arguments array suitable for process launching.
    */
   @SuppressWarnings("null")
   private static String[] prepareCommandLine(String[] cmdlist) 
   {
      // determine if there is a command or not
      int arraySize = (cmdlist == null || cmdlist.length == 0) ? 1 : 3;
      
      // prepare the array by creating a new array with the shell as the
      // first element, then copy the passed array as the subsequent
      // parameters
      String[] list = new String[arraySize];
               
      // WARNING: hard coded to Unix at this time!
      // TODO: move this to the directory and/or provide an alternate
      //       version of this method can allows this to be specified
      //       by the caller         
      list[0] = "sh";         
      
      if (arraySize > 1)
      {
         // build our command back up from parts
         
         StringBuffer sb = new StringBuffer();
         
         for (int i = 0; i < cmdlist.length; i++)
         {
            if (i > 0)
            {
               sb.append(' ');
            }
            if (Quoter.isQuoterCommand(cmdlist[i]))
            {
               String cpath = System.getProperty("java.class.path");
               sb.append("java -cp " + cpath + 
                                     " com.goldencode.p2j.util.Quoter");
            }
            else
            {
               sb.append(cmdlist[i]);
            }
         }
                     
         // pass the command to the shell
         list[1] = "-c";
         list[2] = sb.toString();
      }
      
      return list;
   }
      
   /**
    * Provides a simple mechanism to wait for the exit of a process on an
    * asynchronous basis (on a new thread dedicated to this wait).
    */
   private static class Cleaner
   implements Runnable
   {
     /** The child process identifier for which to wait. */
      private int pid = -1;
      
      /**
       * Constructs an instance that has the data needed to block until the
       * child process exits.
       *
       * @param    pid
       *           Identifies the child process for which to wait.
       */
      public Cleaner(int pid)
      {
         this.pid      = pid;
      }
      
      /**
       * Blocks until the child process exits.
       */
      public void run()
      {
         // block until the child process exits
         LaunchManager.pseudoTerminalWait(pid);
      }
   }
}
